from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .agent import Agent

import asyncio
import datetime
import json
import logging
from .data import DataTypes, load_data, load_file, save_file, save_image
from .utils import format_json

logger = logging.getLogger(__name__)

class Summarizer:
    agent: "Agent"

    def __init__(self, agent):
        self.agent = agent

    async def summarize(self, since_hours, min_responses, batch_size, concurrency, max_retries):
        character_info = self.agent.get_prompt('character.info')
        character_appearance = self.agent.get_prompt('character.appearance')
        image_style = self.agent.get_prompt('character.image_style')
        responses = load_data(
            self.agent.data_dir,
            self.agent.name,
            data_type=DataTypes.MODEL_RESPONSE,
            since=since_hours
        )

        if len(responses) < min_responses:
            raise Exception(f'Not enough responses to summarize ({len(responses)}/{min_responses}).')

        semaphore = asyncio.Semaphore(concurrency)
        batch_summaries = []

        async def process_batch(i, batch):
            async with semaphore:
                for attempt in range(max_retries):
                    try:
                        logger.debug(f'Processing batch {i // batch_size + 1}, attempt {attempt + 1}/{max_retries}')
                        response = await self.agent.get_model_response(
                            'summarize.summarize_one_batch',
                            character_name=self.agent.name,
                            character_info=character_info,
                            data_stream=format_json(batch),
                        )
                        summary = response.get('summary', '')
                        logger.info(f'Finished processing batch {i // batch_size + 1}')
                        logger.debug(f'Batch {i // batch_size + 1} summary:\n{summary}\n--------\n')
                        return summary
                    except Exception as e:
                        if attempt < max_retries - 1:
                            logger.warning(f'Batch {i // batch_size + 1} failed attempt {attempt + 1}: {str(e)}. Retrying...')
                            await asyncio.sleep(5 * (2 ** attempt))
                        else:
                            logger.error(f'Batch {i // batch_size + 1} failed all {max_retries} attempts: {str(e)}')
                            return None

        tasks = []
        for i in range(0, len(responses), batch_size):
            batch = responses[i:i + batch_size]
            task = asyncio.create_task(process_batch(i, batch))
            tasks.append(task)

        logger.info(f'Splitting {len(responses)} responses into {len(tasks)} batches')

        batch_summaries = await asyncio.gather(*tasks, return_exceptions=True)

        successful_summaries = [s for s in batch_summaries if s is not None and not isinstance(s, Exception)]

        if not successful_summaries:
            raise Exception("All batches failed after retries. Cannot generate final summary.")

        for attempt in range(max_retries):
            try:
                logger.debug(f'Processing final summary, attempt {attempt + 1}/{max_retries}')
                final_summary_response = await self.agent.get_model_response(
                    'summarize.summarize_final',
                    character_name=self.agent.name,
                    character_info=character_info,
                    character_appearance=character_appearance,
                    image_style=image_style,
                    data_stream=format_json(successful_summaries),
                )
                logger.info('Finished processing final summary')
                break
            except Exception as e:
                if attempt < max_retries - 1:
                    logger.warning(f'Final summary failed attempt {attempt + 1}: {str(e)}. Retrying...')
                    await asyncio.sleep(5 * (2 ** attempt))
                else:
                    logger.error(f'Final summary failed all {max_retries} attempts: {str(e)}')
                    raise
        
        return final_summary_response

    async def post_story(self, since_hours):
        logger.info(f'Posting story for the last {since_hours} hours')

        logger.info(f'Summarizing data...')
        summary = await self.agent.summarize(since_hours=since_hours)

        logger.info(f'Generating image...')
        prompt = summary.get('image_prompt') + '\n' + self.agent.get_prompt('character.image_style')
        response = await self.agent.model_manager.generate_image(prompt)

        logger.info(f'Saving story...')
        summary_data = {
            'image_prompt': summary.get('image_prompt'),
            'revised_prompt': response.data[0]['revised_prompt'],
            'concise_summary': summary.get('concise_summary'),
        }
        now = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        filename_base = f'{self.agent.data_dir}/{self.agent.name}_{now}'
        save_file(json.dumps(summary_data), f'{filename_base}.json')
        save_image(response.data[0]['url'], f'{filename_base}.png')

        logger.info(f'Posting story...')
        response = await self.agent.api.post_story(summary.get('concise_summary'), response.data[0]['url'])
        logger.info(f'Story posted: {response}')

        logger.info(f'Saving post time...')
        post_data = {
            'last_post_time': now,
        }
        save_file(json.dumps(post_data), self.get_story_filename())

        logger.info(f'Post story done.')

    def time_since_last_post(self):
        try:
            post_data = load_file(self.get_story_filename())
            return datetime.datetime.now() - datetime.datetime.fromisoformat(post_data.get('last_post_time'))
        except Exception as e:
            return datetime.timedelta(weeks=52*42)

    def get_story_filename(self):
        return f'{self.agent.data_dir}/{self.agent.name}_story.json'



import asyncio
import inspect
import json
import logging
from websockets import connect, ConnectionClosedError
from .context import ActionContext
from ..common.const import DEFAULT_WS_ENDPOINT

logger = logging.getLogger(__name__)

class SmartBuilding:
    def __init__(self, api_key, building_id, reconnect_interval=5):
        """
        A smart building in AGIverse.

        :param api_key: Your API key.
        :param building_id: The building ID to connect to.
        :param reconnect_interval: Time in seconds between reconnection attempts.
        """
        self.api_key = api_key
        self.building_id = building_id
        self._event_handlers = {}
        self._action_handlers = {}
        self.players = []
        self.building_info = {}
        self._ws = None
        self._reconnect_interval = reconnect_interval
        self._reconnect = True
        self.set_ws_endpoint(DEFAULT_WS_ENDPOINT)

    def set_ws_endpoint(self, endpoint):
        self.ws_uri = f"{endpoint}?type=building&api-key={self.api_key}&building-id={self.building_id}"

    def event(self, func):
        """
        Decorator to register an event handler.
        """
        self._event_handlers[func.__name__] = func
        return func

    def action(self, action, payload_description='', payment_description=''):
        """
        Decorator to register an action handler.
        """
        def decorator(func):
            self._action_handlers[action] = {
                'func': func,
                'payload_description': payload_description,
                'payment_description': payment_description
            }
            return func
        return decorator

    async def _handle_messages(self):
        while True:
            try:
                message = await self._ws.recv()
                logger.debug(f"Received raw message: {message}")
                try:
                    msg = json.loads(message)
                except json.JSONDecodeError:
                    logger.warning("Received a non-JSON message.")
                    continue
                if not isinstance(msg, dict):
                    logger.warning("Received message is not a dictionary")
                    continue
                msg_type = msg.get("type")
                if msg_type == "building":
                    if not isinstance(msg.get("data"), dict):
                        logger.warning("Received building message with non-dictionary data")
                        continue
                    self.building_info = msg["data"]
                    if 'on_building_info' in self._event_handlers:
                        await self._event_handlers['on_building_info'](self.building_info)
                elif msg_type == "players":
                    if not isinstance(msg.get("data"), list):
                        logger.warning("Received players message with non-list data")
                        continue
                    self.players = msg["data"]
                    if 'on_players' in self._event_handlers:
                        await self._event_handlers['on_players'](self.players)
                elif msg_type == "action":
                    if not isinstance(msg.get("data"), dict):
                        logger.warning("Received action message with non-dictionary data")
                        continue
                    action_name = msg["data"].get("action")
                    action_handler = self._action_handlers.get(action_name)
                    if action_handler:
                        ctx = ActionContext(
                            player_id=msg["data"].get("playerID"),
                            player_name=msg["data"].get("playerName"),
                            building=self,
                            websocket=self._ws,
                            action_id=msg["data"].get("actionID"),
                            action_name=action_name
                        )
                        payload = msg["data"].get("payload")
                        payment = msg["data"].get("payment")

                        num_params = len(inspect.signature(action_handler['func']).parameters)

                        if num_params == 1:
                            await action_handler['func'](ctx)
                        elif num_params == 2:
                            await action_handler['func'](ctx, payload)
                        elif num_params == 3:
                            await action_handler['func'](ctx, payload, payment)
                        else:
                            logger.warning(f"Handler for action '{action_name}' has an unexpected number of parameters")
                    else:
                        logger.warning(f"No handler for action '{action_name}'")
            except ConnectionClosedError:
                logger.warning("Connection closed, attempting to reconnect...")
                break
            except Exception as e:
                logger.error(f"An error occurred: {e}")
                break

    async def _connect(self):
        while self._reconnect:
            try:
                async with connect(self.ws_uri) as ws:
                    self._ws = ws
                    logger.info("WebSocket connection established.")

                    set_actions_message = {
                        "type": "registerActions",
                        "data": {
                            "actions": {
                                action: {
                                    'payload': handler['payload_description'],
                                    'payment': handler['payment_description']
                                }
                                for action, handler in self._action_handlers.items()
                            }
                        }
                    }
                    await self._ws.send(json.dumps(set_actions_message))

                    if 'on_ready' in self._event_handlers:
                        await self._event_handlers['on_ready']()
                    await self._handle_messages()
            except Exception as e:
                logger.error(f"An error occurred: {e}")
                logger.info(f"Reconnecting in {self._reconnect_interval} seconds...")
                await asyncio.sleep(self._reconnect_interval)

    def run(self):
        """
        Starts the connection to the smart building server.
        """
        asyncio.run(self._connect())



import httpx

class API:
    def __init__(self, api_key):
        self.api_key = api_key
        self.api_uri = ""
        self.client = httpx.AsyncClient()

    def set_endpoint(self, endpoint):
        self.api_uri = endpoint

    async def post_story(self, text, image_url=""):
        response = await self.client.post(f"{self.api_uri}/stories", json={
            'apiKey': self.api_key,
            'story': {
                'text': text,
                'imageURL': image_url,
            }
        })
        return response.json()



import logging

logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s - %(name)s - %(levelname)s - %(message)s'
)

import agiverse

building = agiverse.SmartBuilding(api_key="YOUR_API_KEY", building_id="YOUR_BUILDING_ID")

@building.event
async def on_ready():
    logging.info(f"Smart building {building.building_id} is ready to use")

@building.event
async def on_building_info(building_info):
    logging.info(f"Building info: {building_info}")

@building.event
async def on_players(players):
    logging.info(f"Current players in the building: {players}")

@building.action(action="echo", payload_description='{"content": string}')
async def echo(ctx: agiverse.ActionContext, payload):
    if payload and "content" in payload:
        await ctx.send_result(f'You are {ctx.player_name} <{ctx.player_id}>, you said "{payload["content"]}". There are {len(ctx.building.players)} players in the building now.')

@building.action(action="purchase", payload_description='{"content": string}', payment_description='1')
async def purchase(ctx: agiverse.ActionContext, payload, payment):
    # do something
    if payment > 1:
        await ctx.send_result("You are charged $1 for this action!", payment=1)

@building.action(action="withdraw", payload_description='{"content": string}')
async def withdraw(ctx: agiverse.ActionContext, payload, payment):
    # do something
    await ctx.send_result("You are getting paid $1 for this action!", payment=-1)

building.run()



import os
import json
import requests
from datetime import datetime, timedelta

class DataTypes:
    MODEL_RESPONSE = 'model_response'
    SERVER_MESSAGE = 'server_message'
    SYSTEM_MESSAGE = 'system_message'

def load_file(filename):
    with open(filename, 'r') as f:
        return json.load(f)

def save_file(content, filename):
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    with open(filename, 'w') as f:
        f.write(content)

def save_image(url, filename):
    os.makedirs(os.path.dirname(filename), exist_ok=True)
    response = requests.get(url)
    response.raise_for_status()
    with open(filename, 'wb') as f:
        f.write(response.content)

def save_data(data_dir, name, data_type, data):
    os.makedirs(data_dir, exist_ok=True)
    with open(f'{data_dir}/{name}.jsonl', 'a') as f:
        json.dump({
            'type': data_type,
            'data': data,
            'timestamp': datetime.now().isoformat()
        }, f)
        f.write('\n')

def load_data(data_dir, name, data_type=None, since=None, max_responses=None):
    with open(f'{data_dir}/{name}.jsonl', 'r') as f:
        data = [json.loads(line) for line in f if line.strip()]
    
    if data_type is not None:
        data = [d for d in data if d['type'] == data_type]
    
    if since is not None:
        if isinstance(since, (int, float)):
            since_dt = datetime.now() - timedelta(hours=since)
        elif isinstance(since, str):
            since_dt = datetime.fromisoformat(since)
        else:
            since_dt = since
        data = [d for d in data if datetime.fromisoformat(d['timestamp']) >= since_dt]
    
    if max_responses is not None:
        data = data[-max_responses:]

    return data



import asyncio
import logging
from .api import API
from .data import save_data
from .messaging import MessagingHandler
from .model import ModelManager
from .summary import Summarizer
from .utils import load_prompt, load_all_prompts
from ..common.const import DEFAULT_WS_ENDPOINT, DEFAULT_API_ENDPOINT

logger = logging.getLogger(__name__)

class Agent:
    vision_range_buildings = 16
    vision_range_players = 16
    min_num_buildings = 10
    min_num_players = 10
    post_story = True
    post_story_interval_hours = 24

    def __init__(self, api_key, name, data_dir='data'):
        self.api_key = api_key
        self.name = name
        self.data_dir = data_dir
        self._prompts = {}
        self._models = { 'default': 'gpt-4o-mini' }
        self._websocket = None
        self.messaging_handler = MessagingHandler(self)
        self.model_manager = ModelManager(self)
        self.summarizer = Summarizer(self)
        self.working_memory = []
        self.long_term_memory = None
        self.planning = None
        self.api = API(self.api_key)
        self.set_ws_endpoint(DEFAULT_WS_ENDPOINT)
        self.set_api_endpoint(DEFAULT_API_ENDPOINT)
        self._stop_event = asyncio.Event()
        self._tasks = []

    def set_ws_endpoint(self, endpoint):
        self.ws_uri = f"{endpoint}?type=player&api-key={self.api_key}&name={self.name}"

    def set_api_endpoint(self, endpoint):
        self.api.set_endpoint(endpoint)

    def set_chat_completion_function(self, f):
        self.model_manager.set_chat_completion_function(f)

    def set_image_generation_function(self, f):
        self.model_manager.set_image_generation_function(f)

    async def summarize(self, since_hours=24, min_responses=10, batch_size=256, concurrency=8, max_retries=4):
        return await self.summarizer.summarize(since_hours, min_responses, batch_size, concurrency, max_retries)

    def get_all_prompts(self):
        """
        Get all prompts used by combining default prompts with any custom prompts.

        Returns:
            dict: Combined dictionary of all prompts
        """
        prompts = load_all_prompts()
        prompts.update(self._prompts)
        return prompts

    def get_prompt(self, prompt_key):
        """
        Get a specific prompt by key, first checking custom prompts then default prompts.

        Args:
            prompt_key (str): Key of the desired prompt

        Returns:
            str: The prompt text
        """
        prompt = self._prompts.get(prompt_key)
        if not prompt:
            prompt = load_prompt(prompt_key)
        return prompt

    def set_prompt(self, prompt_key, prompt):
        """
        Set a custom prompt. Set to None or empty string to use default prompt.
        The prompt will be immediately effective even if agent is already running.

        Args:
            prompt_key (str): Key of the prompt
            prompt (str): The prompt text to set
        """
        self._prompts[prompt_key] = prompt

    def set_prompts(self, prompts):
        """
        Replace all custom prompts with a new dictionary.
        Missing keys from the provided dictionary will be loaded from default prompts.
        The prompts will be immediately effective even if agent is already running.

        Args:
            prompts (dict): Dictionary of prompts to set
        """
        self._prompts = prompts

    def update_prompts(self, prompts):
        """
        Update custom prompts by merging with provided prompts.
        The prompts will be immediately effective even if agent is already running.

        Args:
            prompts (dict): Dictionary of prompts to merge with existing prompts
        """
        self._prompts.update(prompts)

    def get_model(self, prompt_key):
        """
        Get the model for a specific prompt key, falling back to default model if not found.

        Args:
            prompt_key (str): Key to look up model for

        Returns:
            str: The model name to use
        """
        return self._models.get(prompt_key, self._models.get('default'))

    def set_model(self, prompt_key, model):
        """
        Set the model to use for a specific prompt key.
        Set the model of key 'default' to set the default model.
        The model will be immediately effective even if agent is already running.

        Args:
            prompt_key (str): Key to set model for
            model (str): Name of the model to use
        """
        self._models[prompt_key] = model

    async def get_model_response(self, prompt_key, **kwargs):
        prompt = self.get_prompt(prompt_key).format(**kwargs)
        return await self.model_manager.get_model_response(prompt, prompt_key=prompt_key)

    def set_models(self, models):
        """
        Replace all custom models with a new dictionary.
        The model of key 'default' will be used as the default model.
        The models will be immediately effective even if agent is already running.

        Args:
            models (dict): Dictionary mapping prompt keys to model names
        """
        self._models = models

    def update_models(self, models):
        """
        Update custom models by merging with provided models.
        The model of key 'default' will be used as the default model.
        The models will be immediately effective even if agent is already running.

        Args:
            models (dict): Dictionary of models to merge with existing models
        """
        self._models.update(models)

    async def _start_messaging(self):
        while not self._stop_event.is_set():
            try:
                async with await self.messaging_handler.connect(self.ws_uri) as websocket:
                    self._websocket = websocket
                    await self.messaging_handler.handle_messages(websocket, self._stop_event)
            except asyncio.CancelledError:
                logger.info("Messaging task cancelled.")
                break
            except Exception as e:
                logger.error(f"Error: {e}. Reconnecting in 5 seconds...")
                await asyncio.sleep(5)

    async def _start_summarizer(self):
        if not self.post_story:
            return
        while not self._stop_event.is_set():
            await asyncio.sleep(60)
            try:
                time_since_last_post = self.summarizer.time_since_last_post()
                if time_since_last_post.total_seconds() >= self.post_story_interval_hours * 3600:
                    await self.summarizer.post_story(since_hours=self.post_story_interval_hours)
            except asyncio.CancelledError:
                logger.info("Summarizer task cancelled.")
                break
            except Exception as e:
                logger.error(f"Post story failed: {e}")

    def _save_data(self, data_type, data):
        save_data(self.data_dir, self.name, data_type, data)

    def run(self):
        """
        Synchronous entry point to run the agent.
        """
        try:
            asyncio.run(self.start())
        except KeyboardInterrupt:
            logger.info("Agent interrupted by user. Shutting down...")
            asyncio.run(self.stop())

    async def start(self):
        """
        Asynchronous entry point to run the agent.
        """
        self._tasks = [
            asyncio.create_task(self._start_messaging()),
            asyncio.create_task(self._start_summarizer())
        ]
        await self._stop_event.wait()
        for task in self._tasks:
            task.cancel()
        await asyncio.gather(*self._tasks, return_exceptions=True)
        logger.info("Agent has been stopped.")

    async def stop(self):
        """
        Stop the agent.
        """
        logger.info("Stopping the agent...")
        self._stop_event.set()
        if self._websocket:
            await self._websocket.close()
        if hasattr(self, '_tasks'):
            await asyncio.gather(*self._tasks, return_exceptions=True)



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .agent import Agent

import asyncio
import json
import logging
import time
import os
import websockets
from datetime import datetime
from .data import DataTypes
from .utils import (
    distance_obj,
    is_valid_state_data,
    is_valid_map_data,
    is_valid_player_data,
    remove_additional_data,
    seconds_since,
    format_json,
    format_memory,
)

logger = logging.getLogger(__name__)

class MessagingHandler:
    agent: "Agent"

    def __init__(self, agent):
        self.agent = agent
        self.message_queue = asyncio.Queue()
        self.map_data = None
        self.nearby_map = None
        self.players_data = None
        self.nearby_players = None
        self.assets_data = None
        self.inventory_data = None
        self.state_data = None
        self.available_actions = None
        self.system_messages = []
        self.other_data = []
        self.last_prompt_time = None
        self.use_model = False
        self.min_model_interval = int(os.getenv('MIN_MODEL_INTERVAL', 5))
        self.max_model_interval = int(os.getenv('MAX_MODEL_INTERVAL', 60))
        self.model_interval = self.min_model_interval
        self.action_id = int(time.time() * 1000)

    async def connect(self, uri):
        self.websocket = await websockets.connect(uri)
        return self

    async def __aenter__(self):
        return self.websocket

    async def __aexit__(self, exc_type, exc, tb):
        await self.websocket.close()

    async def handle_messages(self, websocket, stop_event):
        receive_task = asyncio.create_task(self._receive_messages(websocket, stop_event))
        process_task = asyncio.create_task(self._process_messages(websocket, stop_event))

        done, pending = await asyncio.wait(
            [receive_task, process_task],
            return_when=asyncio.FIRST_COMPLETED,
        )

        for task in done:
            if task.exception():
                raise task.exception()

        for task in pending:
            task.cancel()
            try:
                await task
            except asyncio.CancelledError:
                logger.info(f"Task {task} has been cancelled successfully.")

    async def _receive_messages(self, websocket, stop_event):
        try:
            async for message in websocket:
                if stop_event.is_set():
                    logger.info("Stop event detected. Exiting handle_messages.")
                    break
                await self.message_queue.put(json.loads(message))
        except websockets.exceptions.ConnectionClosed as e:
            logger.info(f"Receive messages - Connection closed: {e}")
            raise

    async def _process_messages(self, websocket, stop_event):
        try:
            while True:
                if stop_event.is_set():
                    logger.info("Stop event detected. Exiting handle_messages.")
                    break

                try:
                    message = await asyncio.wait_for(self.message_queue.get(), timeout=60)
                    if stop_event.is_set():
                        logger.info("Stop event detected. Exiting handle_messages.")
                        break
                    if not await self._process_message(message):
                        continue
                except asyncio.TimeoutError:
                    self.use_model = True
                    message = None
                except asyncio.CancelledError:
                    raise

                if not self.message_queue.empty():
                    continue

                if len(self.system_messages) == 0:
                    seconds_since_last_prompt = seconds_since(self.last_prompt_time) if self.last_prompt_time else self.max_model_interval
                    if seconds_since_last_prompt < self.model_interval:
                        continue
                    if seconds_since_last_prompt < self.max_model_interval and not self.use_model:
                        continue

                try:
                    response = await self._generate_model_response(websocket)
                except asyncio.CancelledError:
                    raise
                except Exception as e:
                    logger.error(f"Error generating model response: {e}")
                    continue

                try:
                    await self._process_model_response(response)
                except asyncio.CancelledError:
                    raise
                except Exception as e:
                    logger.error(f"Error processing model response: {e}")
                    continue

                try:
                    await self._send_action(response, websocket)
                except asyncio.CancelledError:
                    raise
                except Exception as e:
                    logger.error(f"Error sending action: {e}")
                    raise
        except asyncio.CancelledError:
            logger.info("Process messages has been cancelled.")
            raise
        except Exception as e:
            logger.error(f"Process messages encountered an unexpected error: {e}")
            raise

    async def _process_message(self, message):
        msg_type = message.get('type')
        data = message.get('data')

        if msg_type != 'tickEnd':
            logger.info(f"Received: {msg_type} {data}")

        if msg_type == 'map':
            self.map_data = data
            self.use_model = True
        elif msg_type == 'players':
            self.players_data = data
        elif msg_type == 'assets':
            self.assets_data = data
            self.use_model = True
        elif msg_type == 'inventory':
            self.inventory_data = data
            self.use_model = True
        elif msg_type == 'state':
            self.state_data = data
            if self.state_data['state'] in ['moving', 'learning', 'working']:
                self.model_interval = self.max_model_interval if self.state_data['state'] != 'moving' else self.model_interval
            else:
                self.model_interval = self.min_model_interval
            self.use_model = True
        elif msg_type == 'availableActions':
            self.available_actions = data
            self.use_model = True
        elif msg_type == 'tickEnd':
            pass
        elif msg_type == 'system':
            self.system_messages.append(message)
            self.use_model = True
            self.agent._save_data(DataTypes.SYSTEM_MESSAGE, message)
        else:
            self.other_data.append(message)
            self.use_model = True
            self.agent._save_data(DataTypes.SERVER_MESSAGE, message)

        if is_valid_state_data(self.state_data) and is_valid_map_data(self.map_data, 'buildings'):
            nearby_map = {'buildings': [], 'my_houses': []}
            for building in self.map_data['buildings']:
                building['distance'] = distance_obj(building['entrance'], self.state_data)

            sorted_buildings = sorted(
                self.map_data['buildings'],
                key=lambda b: b.get('distance', float('inf'))
            )

            for building in sorted_buildings:
                remove_additional_data(building)
                distance = building.get('distance', float('inf'))
                if distance > self.agent.vision_range_buildings and len(nearby_map['buildings']) >= self.agent.min_num_buildings:
                    break
                nearby_map['buildings'].append(building)

            if is_valid_map_data(self.assets_data, 'rentedBuildings'):
                for building in self.assets_data['rentedBuildings']:
                    building['distance'] = distance_obj(building['entrance'], self.state_data)

                sorted_rented_buildings = sorted(
                    self.assets_data['rentedBuildings'],
                    key=lambda b: b.get('distance', float('inf'))
                )

                for building in sorted_rented_buildings:
                    remove_additional_data(building)
                    distance = building.get('distance', float('inf'))
                    if distance > self.agent.vision_range_buildings and len(nearby_map['my_houses']) >= self.agent.min_num_buildings:
                        break
                    nearby_map['my_houses'].append(building)

            self.nearby_map = nearby_map

        if is_valid_state_data(self.state_data) and is_valid_player_data(self.players_data):
            nearby_players = []
            for player in self.players_data:
                player['distance'] = distance_obj(player, self.state_data)

            sorted_players = sorted(
                self.players_data,
                key=lambda p: p.get('distance', float('inf'))
            )

            for player in sorted_players:
                if player.get('playerID') == self.state_data.get('playerID'):
                    continue
                distance = player.get('distance', float('inf'))
                if distance > self.agent.vision_range_players and len(nearby_players) >= self.agent.min_num_players:
                    break
                nearby_players.append(player)

            self.nearby_players = nearby_players

        return msg_type == 'tickEnd'

    async def _generate_model_response(self, websocket):
        return await self.agent.get_model_response(
            'agent.agent',
            character_name=self.agent.name,
            character_info=self.agent.get_prompt('character.info'),
            nearby_map=format_json(self.nearby_map),
            nearby_players=format_json(self.nearby_players),
            assets=format_json(self.assets_data),
            inventory=format_json(self.inventory_data),
            state=format_json(self.state_data),
            available_actions=format_json(self.available_actions),
            working_memory=format_memory(self.agent.working_memory),
            long_term_memory=self.agent.long_term_memory,
            planning=self.agent.planning,
            system_messages=format_json(self.system_messages),
            messages=format_json(self.other_data),
            current_time=datetime.now().strftime('%Y-%m-%d %H:%M:%S'),
        )

    async def _process_model_response(self, response):
        self.agent._save_data(DataTypes.MODEL_RESPONSE, response)

        logger.info('-' * 100)
        logger.info(f"Observation: {response['observation']}")
        logger.info(f"Thought: {response['thought']}")
        logger.info(f"Long term memory: {response.get('longTermMemory', self.agent.long_term_memory)}")
        logger.info(f"Planning: {response.get('planning', self.agent.planning)}")
        logger.info(f"Action: {response['action']}")
        logger.info(f"System message reply action: {response.get('systemMessageReplyAction')}")
        logger.info('-' * 100)

        new_long_term_memory = response.get('longTermMemory', '')
        if new_long_term_memory:
            self.agent.long_term_memory = new_long_term_memory

        new_planning = response.get('planning', '')
        if new_planning:
            self.agent.planning = new_planning

        self.agent.working_memory.append(response)
        if len(self.agent.working_memory) > 10:
            self.agent.working_memory.pop(0)

        self.system_messages = []
        self.other_data = []
        self.last_prompt_time = datetime.now()
        self.use_model = False
        self.model_interval = self.min_model_interval

    async def _send_action(self, response, websocket):
        if not response.get('action') or response['action'].get('action') == 'no action':
            logger.info('No action to perform.')
        else:
            response['action']['actionID'] = self.action_id
            self.action_id += 1
            logger.info(f"Sending action: {response['action']}")
            await websocket.send(json.dumps(response['action']))

        if response.get('systemMessageReplyAction'):
            logger.info(f"Sending system message reply action: {response['systemMessageReplyAction']}")
            await websocket.send(json.dumps(response['systemMessageReplyAction']))



import agiverse

agent = agiverse.Agent(api_key="Your API Key", name="Your character name")

character_info = agent.get_prompt("character.info")
character_info += "\nYou are a resourceful person who values both success and helping others."
agent.set_prompt("character.info", character_info)

agent.run()



import json

class ActionContext:
    def __init__(self, player_id, player_name, building, websocket, action_id, action_name):
        self.player_id = player_id
        self.player_name = player_name
        self.building = building
        self.websocket = websocket
        self.action_id = action_id
        self.action_name = action_name

    async def send_result(self, payload, payment=0):
        """
        Sends the result of the action back to the server.

        :param payload: The payload to send back as the result.
        """
        action_result = {
            "type": "actionResult",
            "data": {
                "playerID": self.player_id,
                "action": self.action_name,
                "actionID": self.action_id,
                "payload": payload,
                "payment": payment,
            }
        }
        await self.websocket.send(json.dumps(action_result))



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .agent import Agent

import asyncio
import datetime
import json
import logging
import os
import litellm

logger = logging.getLogger(__name__)

litellm.drop_params = True

class ModelManager:
    agent: "Agent"

    def __init__(self, agent):
        self.agent = agent
        self.usage_history = []
        self.max_history_hours = 24
        self.chat_completion = litellm.acompletion
        self.image_generation = litellm.aimage_generation

    def set_chat_completion_function(self, f):
        self.chat_completion = f

    def set_image_generation_function(self, f):
        self.image_generation = f

    async def get_model_response(self, prompt, prompt_key='default'):
        response = await asyncio.wait_for(
            self.chat_completion(
                model=self.agent.get_model(prompt_key),
                messages=[{"role": "user", "content": prompt}],
                response_format={"type": "json_object"},
            ),
            timeout=os.getenv('MODEL_TIMEOUT', 60),
        )

        try:
            logger.info(f'Input tokens: {response.usage.prompt_tokens}, output tokens: {response.usage.completion_tokens}')
            current_time = datetime.datetime.now()
            self.usage_history.append((current_time, response.usage.prompt_tokens, response.usage.completion_tokens))
            cutoff_time = current_time - datetime.timedelta(hours=self.max_history_hours)
            self.usage_history = [stat for stat in self.usage_history if stat[0] > cutoff_time]
            stats = self.get_usage_stats(hours=1)
            logger.info(f'Last hour input tokens: {stats["input_tokens"]}, output tokens: {stats["output_tokens"]}')
            stats = self.get_usage_stats(hours=24)
            logger.info(f'Last 24 hours input tokens: {stats["input_tokens"]}, output tokens: {stats["output_tokens"]}')
        except Exception as e:
            logger.error(f'Error updating usage stats: {e}')
        
        content = response.choices[0].message.content
        if content.startswith('```json') and content.endswith('```'):
            content = content[7:-3]
        return json.loads(content)
    
    def get_usage_stats(self, hours=None):
        """Get usage statistics for the specified number of hours."""
        if not self.usage_history:
            return {'input_tokens': 0, 'output_tokens': 0}

        if hours:
            cutoff_time = datetime.datetime.now() - datetime.timedelta(hours=hours)
            filtered_stats = [stat for stat in self.usage_history if stat[0] > cutoff_time]
        else:
            filtered_stats = self.usage_history
        
        return {
            'input_tokens': sum(stat[1] for stat in filtered_stats),
            'output_tokens': sum(stat[2] for stat in filtered_stats)
        }

    async def generate_image(self, prompt, max_retries=5):
        for attempt in range(max_retries):
            try:
                response = await asyncio.wait_for(
                    self.image_generation(
                        model='dall-e-3',
                        quality='hd',
                        size='1792x1024',
                        prompt=prompt,
                    ),
                    timeout=os.getenv('IMAGE_GENERATION_TIMEOUT', 120),
                )
                return response
            except Exception as e:
                if attempt == max_retries - 1:
                    raise
                print(f"Attempt {attempt + 1} failed. Retrying...")
                await asyncio.sleep(5 * (2 ** attempt))


import datetime
import os
import yaml

def load_prompt(prompt_path):
    parts = prompt_path.split('.')
    prompt_name = parts[-1]
    file_name = '.'.join(parts[:-1])
    prompts = load_prompt_file(file_name)
    return prompts.get(prompt_name, '')

def load_prompt_file(file_name):
    script_dir = os.path.dirname(os.path.abspath(__file__))
    prompts_path = os.path.join(script_dir, 'prompts', f'{file_name}.yaml')
    with open(prompts_path, 'r', encoding='utf-8') as file:
        return yaml.safe_load(file)

def load_all_prompts():
    """
    Reads all YAML files in the prompts directory using the load_prompt_file function,
    loads their contents, and merges them into a single dictionary. The keys in the resulting
    dictionary are formatted as 'filename.key_in_file'.

    Returns:
        dict: A merged dictionary containing all key-value pairs from the YAML files.
    """
    script_dir = os.path.dirname(os.path.abspath(__file__))
    prompts_path = os.path.join(script_dir, 'prompts')
    all_prompts = {}
    
    for filename in os.listdir(prompts_path):
        if filename.endswith('.yaml'):
            file_name = os.path.splitext(filename)[0]
            try:
                data = load_prompt_file(os.path.join(prompts_path, file_name))
                if isinstance(data, dict):
                    for key, value in data.items():
                        all_prompts[f"{file_name}.{key}"] = value
                else:
                    print(f"Warning: {filename} does not contain a valid YAML dictionary.")
            except Exception as e:
                print(f"Error reading {filename}: {e}")
    
    return all_prompts

def format_json(json_obj, indent=0):
    indent_str = '    ' * indent
    formatted_string = ""
    if isinstance(json_obj, dict):
        for key, value in json_obj.items():
            formatted_string += f"{indent_str}{key}:"
            if isinstance(value, (dict, list)):
                formatted_string += "\n" + format_json(value, indent + 1)
            else:
                formatted_string += f" {value}\n"
    elif isinstance(json_obj, list):
        for item in json_obj:
            formatted_string += indent_str + "- "
            if isinstance(item, (dict, list)):
                formatted_string += "\n" + format_json(item, indent + 1)
            else:
                formatted_string += f"{item}\n"
    else:
        formatted_string += f"{indent_str}{json_obj}\n"
    return formatted_string

def format_memory(memory):
    memory_str = ''
    for entry in memory:
        memory_str += f"Observation: {entry.get('observation', '')}\n"
        memory_str += f"Thought: {entry.get('thought', '')}\n"
        memory_str += f"Action: {format_json(entry.get('action', {}))}\n\n"
    return memory_str

def remove_additional_data(building):
    for key in ['boundary', 'rent']:
        building.pop(key, None)

def distance(x1, y1, x2, y2):
    return abs(x1 - x2) + abs(y1 - y2)

def distance_obj(obj1, obj2):
    if 'x' in obj1 and 'y' in obj1:
        x1, y1 = obj1['x'], obj1['y']
    elif 'locationX' in obj1 and 'locationY' in obj1:
        x1, y1 = obj1['locationX'], obj1['locationY']
    else:
        x1, y1 = 0, 0

    if 'x' in obj2 and 'y' in obj2:
        x2, y2 = obj2['x'], obj2['y']
    elif 'locationX' in obj2 and 'locationY' in obj2:
        x2, y2 = obj2['locationX'], obj2['locationY']
    else:
        x2, y2 = 0, 0

    return distance(x1, y1, x2, y2)

def is_valid_state_data(state_data):
    return (
        state_data 
        and isinstance(state_data, dict) 
        and 'playerID' in state_data
        and 'locationX' in state_data
        and 'locationY' in state_data
    )

def is_valid_player_data(players_data):
    return (
        players_data 
        and isinstance(players_data, list)
    )

def is_valid_map_data(map_data, key):
    return (
        map_data 
        and isinstance(map_data, dict) 
        and key in map_data 
        and isinstance(map_data[key], list)
    )

def seconds_since(time):
    return (datetime.datetime.now() - time).total_seconds()



import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

if not logger.hasHandlers():
    handler = logging.StreamHandler()
    formatter = logging.Formatter(
        f'%(asctime)s - {__name__} - %(levelname)s - %(message)s'
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)

from .agent import Agent
from .smart_building import ActionContext, SmartBuilding

__all__ = ['Agent', 'ActionContext', 'SmartBuilding']
